@@ -518,6 +518,8 @@ func (e *extractor) extractPkg(root string, p *packages.Package) error {
 		}
 
 		addDoc(f.Doc, cuePkg)
+		lenPreamble := len(cueFile.Decls)
+
 		for _, d := range f.Decls {
 			switch d := d.(type) {
 			case *ast.GenDecl:
@@ -525,8 +527,9 @@ func (e *extractor) extractPkg(root string, p *packages.Package) error {
 			}
 		}
 
-		if outFile == "" && len(cueFile.Decls) == 0 && f.Doc == nil {
+		if outFile == "" && len(cueFile.Decls) == lenPreamble && f.Doc == nil {
 			// By default, empty files are not generated.
+			// Note that cueFile.Decls is never empty, as we add top-level docs and a package clause.
 			continue
 		}
 
@@ -659,6 +662,9 @@ func (e *extractor) recordConsts(x *ast.GenDecl) {
 			continue
 		}
 		for _, n := range v.Names {
+			if n.Name == "_" {
+				continue
+			}
 			typ := e.pkg.TypesInfo.TypeOf(n).String()
 			e.consts[typ] = append(e.consts[typ], n.Name)
 		}
@@ -746,9 +752,6 @@ func (e *extractor) reportDecl(x *ast.GenDecl) (a []cueast.Decl) {
 				var exprs []cueast.Expr
 				var named []cueast.Decl
 				for _, v := range enums {
-					if v == "_" {
-						continue
-					}
 					label := cueast.NewString(v)
 					cueast.SetRelPos(label, cuetoken.Blank)
 
