@@ -15,7 +15,6 @@
 package cue_test
 
 import (
-	"bytes"
 	"fmt"
 	"io"
 	"math"
@@ -25,11 +24,13 @@ import (
 	"strings"
 	"testing"
 
+	"github.com/go-quicktest/qt"
 	"github.com/google/go-cmp/cmp"
 
 	"cuelang.org/go/cue"
 	"cuelang.org/go/cue/ast"
 	"cuelang.org/go/cue/cuecontext"
+	"cuelang.org/go/cue/errors"
 	"cuelang.org/go/internal/astinternal"
 	"cuelang.org/go/internal/core/adt"
 	"cuelang.org/go/internal/core/debug"
@@ -57,8 +58,22 @@ func TestAPI(t *testing.T) {
 		input string
 		fun   func(i cue.Value) cue.Value
 		want  string
-		skip  bool
 	}{{
+		// Issue #567
+		input: `
+		#runSpec: {action: int}
+
+		v: {ction: 1}
+				`,
+		fun: func(i cue.Value) cue.Value {
+			runSpec := i.LookupDef("#runSpec")
+			v := i.Lookup("v")
+
+			res := runSpec.Unify(v)
+			return res
+		},
+		want: "_|_ // #runSpec.ction: field not allowed",
+	}, {
 		// Issue #567
 		input: `
 		#runSpec: {action: foo: int}
@@ -72,8 +87,6 @@ func TestAPI(t *testing.T) {
 			return res
 		},
 		want: "_|_ // #runSpec.ction: field not allowed",
-
-		skip: true,
 	}, {
 		// Issue #567
 		input: `
@@ -134,13 +147,10 @@ func TestAPI(t *testing.T) {
 		},
 		want: `"test1"`,
 	}}
-	for _, tc := range testCases {
-		if tc.skip {
-			continue
-		}
-		cuetdtest.FullMatrix.Run(t, "", func(t *testing.T, m *cuetdtest.M) {
-			m.TODO_V3(t) // P1: faulty closedness
-
+	for i, tc := range testCases {
+		_ = i // for debugging purposes
+		m := cuetdtest.FullMatrix
+		m.Run(t, "", func(t *testing.T, m *cuetdtest.M) {
 			ctx := m.CueContext()
 
 			valIn := mustCompile(t, ctx, tc.input)
@@ -268,13 +278,11 @@ func TestValueType(t *testing.T) {
 		kind:           cue.StructKind,
 		incompleteKind: cue.StructKind,
 		concrete:       true,
-		closed:         true,
 	}, {
 		value:          `v: []`,
 		kind:           cue.ListKind,
 		incompleteKind: cue.ListKind,
 		concrete:       true,
-		closed:         true,
 	}, {
 		value:          `v: [...int]`,
 		kind:           cue.ListKind,
@@ -331,16 +339,12 @@ func TestValueType(t *testing.T) {
 		kind:           cue.StructKind,
 		incompleteKind: cue.StructKind,
 		concrete:       true,
-		closed:         true,
 	}, {
 		value: `v: #Foo | int
 		#Foo: {
 			name: string,
 			}`,
 		incompleteKind: cue.StructKind | cue.IntKind,
-		// Hard to tell what is correct here, but For backwards compatibility,
-		// this is false.
-		closed: false,
 	}}
 	for _, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, tc.value, func(t *testing.T, m *cuetdtest.M) {
@@ -356,9 +360,6 @@ func TestValueType(t *testing.T) {
 			if got := v.IsConcrete(); got != tc.concrete {
 				t.Errorf("IsConcrete: got %v; want %v", got, tc.concrete)
 			}
-			if got := v.IsClosed(); got != tc.closed {
-				t.Errorf("IsClosed: got %v; want %v", got, tc.closed)
-			}
 		})
 	}
 }
@@ -432,7 +433,7 @@ func TestInt(t *testing.T) {
 
 			vu, err := n.Uint64()
 			checkErr(t, err, tc.errU, "Uint64")
-			if vu != uint64(tc.uint) {
+			if vu != tc.uint {
 				t.Errorf("Uint64: got %v; want %v", vu, tc.uint)
 			}
 		})
@@ -580,6 +581,74 @@ func TestFloat(t *testing.T) {
 	}
 }
 
+func TestBigFloat(t *testing.T) {
+	testCases := []struct {
+		value    string
+		bigFloat *big.Float
+		kind     cue.Kind
+		err      string
+	}{{
+		value:    "1",
+		bigFloat: bigFloat("1"),
+		kind:     cue.IntKind,
+	}, {
+		value:    "-1",
+		bigFloat: bigFloat("-1"),
+		kind:     cue.IntKind,
+	}, {
+		value:    "0.0",
+		bigFloat: bigFloat("0.0"),
+		kind:     cue.FloatKind,
+	}, {
+		value:    "1.0",
+		bigFloat: bigFloat("1.0"),
+		kind:     cue.FloatKind,
+	}, {
+		value:    "2.6",
+		bigFloat: bigFloat("2.6"),
+		kind:     cue.FloatKind,
+	}, {
+		value:    "20.600",
+		bigFloat: bigFloat("20.600"),
+		kind:     cue.FloatKind,
+	}, {
+		value:    "1/0",
+		bigFloat: nil,
+		err:      "division by zero",
+		kind:     cue.BottomKind,
+	}, {
+		value:    "1.797693134862315708145274237317043567982e+308",
+		bigFloat: bigFloat("1.797693134862315708145274237317043567982e+308"),
+		kind:     cue.FloatKind,
+	}, {
+		value:    "-1.797693134862315708145274237317043567982e+308",
+		bigFloat: bigFloat("-1.797693134862315708145274237317043567982e+308"),
+		kind:     cue.FloatKind,
+	}, {
+		value:    "4.940656458412465441765687928682213723650e-324",
+		bigFloat: bigFloat("4.940656458412465441765687928682213723650e-324"),
+		kind:     cue.FloatKind,
+	}, {
+		value:    "-4.940656458412465441765687928682213723650e-324",
+		bigFloat: bigFloat("-4.940656458412465441765687928682213723650e-324"),
+		kind:     cue.FloatKind,
+	}}
+	for _, tc := range testCases {
+		cuetdtest.FullMatrix.Run(t, tc.value, func(t *testing.T, m *cuetdtest.M) {
+			n := getValue(m, tc.value)
+			if n.Kind() != tc.kind {
+				t.Fatal("Not a number")
+			}
+
+			bf, err := n.Float(nil)
+			checkErr(t, err, tc.err, "Float")
+			if bf != nil && bf.Cmp(tc.bigFloat) != 0 {
+				t.Errorf("Float: got %v; want %v", bf, tc.bigFloat)
+			}
+		})
+	}
+}
+
 func TestString(t *testing.T) {
 	testCases := []struct {
 		value string
@@ -762,8 +831,6 @@ func TestFields(t *testing.T) {
 		err   string
 		opts  []cue.Option
 		path  string
-
-		todoV3 bool
 	}{{
 		value: `{ #def: 1, _hidden: 2, opt?: 3, reg: 4 }`,
 		res:   "{reg:4,}",
@@ -794,7 +861,7 @@ func TestFields(t *testing.T) {
 		res:   "{reg:4,}",
 	}, {
 		value: `{a:1,b:2,c:int}`,
-		err:   "cannot convert incomplete value",
+		res:   "{a:1,b:2,c:int,}",
 	}, {
 		value: `
 			step1: {}
@@ -803,7 +870,7 @@ func TestFields(t *testing.T) {
 			   step3: {prefix: step2.value}
 			}
 			_hidden: 3`,
-		res: `{step1:{},step2:{"prefix":3},}`,
+		res: `{step1:,step2:prefix: 3,}`,
 	}, {
 		opts: []cue.Option{cue.Final()},
 		value: `
@@ -837,9 +904,37 @@ func TestFields(t *testing.T) {
 		value: `a: x, x: y, y: b: 1`,
 		path:  "a",
 		res:   `{b:1,}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `[=~"^a"]: 1`,
+		res:   `{[=~"^a"]:1,}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `[=~"^a"]: 1, [string]: 2`,
+		res:   `{[=~"^a"]:1,[string]:2,}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `[=~"^a"]: >1, a: 2`,
+		res:   `{[=~"^a"]:>1,a:2,}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `{[_]: 1}`,
+		res:   `{[_]:1,}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `{...}`,
+		res:   `{}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `{...}`,
+		res:   `{}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `{[1, 2, ...int], [string]: int}`,
+		res:   `{[>=2]:int,[string]:int,0:1,1:2,}`,
 	}}
 	for _, tc := range testCases {
-		cuetdtest.DevOnlyMatrix.Run(t, tc.value, func(t *testing.T, m *cuetdtest.M) {
+		cuetdtest.SmallMatrix.Run(t, "", func(t *testing.T, m *cuetdtest.M) {
 			f := getValue(m, tc.value)
 
 			obj := f.LookupPath(cue.ParsePath(tc.path))
@@ -849,11 +944,23 @@ func TestFields(t *testing.T) {
 
 			buf := []byte{'{'}
 			for iter.Next() {
-				buf = append(buf, iter.Selector().String()...)
+				sel := iter.Selector()
+				if sel.ConstraintType() == cue.PatternConstraint {
+					qt.Check(t, qt.Equals(
+						sel.String(),
+						fmt.Sprintf("[%#v]", sel.Pattern()),
+					))
+					wantType := cue.StringLabel | cue.PatternConstraint
+					if obj.Kind() == cue.ListKind {
+						wantType = cue.IndexLabel | cue.PatternConstraint
+					}
+					qt.Check(t, qt.Equals(sel.Type(), wantType))
+				}
+				buf = append(buf, sel.String()...)
 				buf = append(buf, ':')
-				b, err := iter.Value().MarshalJSON()
-				checkFatal(t, err, tc.err, "Obj.At")
-				buf = append(buf, b...)
+				v := iter.Value()
+				checkFatal(t, v.Err(), tc.err, "Obj.At")
+				buf = fmt.Appendf(buf, "%#v", v)
 				buf = append(buf, ',')
 			}
 			buf = append(buf, '}')
@@ -863,14 +970,20 @@ func TestFields(t *testing.T) {
 
 			iter, _ = obj.Fields(tc.opts...)
 			for iter.Next() {
-				want, err := iter.Value().MarshalJSON()
-				checkFatal(t, err, tc.err, "Obj.At2")
+				v := iter.Value()
+				checkFatal(t, v.Err(), tc.err, "Obj.At2")
+				want := fmt.Sprintf("%#v", v)
 
-				got, err := obj.LookupPath(cue.MakePath(iter.Selector())).MarshalJSON()
-				checkFatal(t, err, tc.err, "Obj.At2")
-
-				if !bytes.Equal(got, want) {
-					t.Errorf("Lookup: got %q; want %q", got, want)
+				got := obj.LookupPath(cue.MakePath(iter.Selector()))
+				if iter.FieldType().ConstraintType() == cue.PatternConstraint {
+					// Can't look up iterated pattern constraints.
+					qt.Assert(t, qt.ErrorMatches(
+						got.Err(),
+						`.*cannot look up pattern constraints.*`,
+					))
+				} else {
+					checkFatal(t, err, tc.err, "Obj.At2")
+					qt.Assert(t, qt.Equals(fmt.Sprintf("%#v", got), want))
 				}
 			}
 			v := obj.LookupPath(cue.MakePath(cue.Str("non-existing")))
@@ -1379,6 +1492,21 @@ func TestFillPath(t *testing.T) {
 			x:    1,
 			path: cue.MakePath(cue.Str("b").Optional()),
 			out:  `{b: 1}`,
+		}, {
+			in: `
+				_schema: _
+				_schema
+				apiVersion: "foo"
+				spec: group: "foo"
+			`,
+			x: ctx.CompileString(`
+				#Schema: spec!: group!: string
+			`).LookupPath(cue.MakePath(cue.Def("#Schema"))),
+			path: cue.MakePath(cue.Hid("_schema", "_")),
+			out: `
+				apiVersion: "foo"
+				spec: group: "foo"
+            `,
 		}}
 
 		for _, tc := range testCases {
@@ -1434,8 +1562,6 @@ func TestAllows(t *testing.T) {
 		in    string
 		sel   cue.Selector
 		allow bool
-
-		todo_nosharing bool
 	}{{
 		desc: "allow new field in open struct",
 		in: `
@@ -1556,17 +1682,17 @@ func TestAllows(t *testing.T) {
 			x: 1
 			`,
 		sel: cue.AnyString,
-
-		// TODO(v0.6.0)
-		// }, {
-		// 	desc: "definition in closed scalar",
-		// 	in: `
-		// 	x: #Def
-		// 	x: 1
-		// 	#Def: int
-		// 	`,
-		// 	sel:   cue.AnyDefinition,
-		// 	allow: true,
+	}, {
+		// TODO: note that this may be affected by https://cuelang.org/issue/543
+		// if or when we decide to implement it in the future.
+		desc: "definition in closed scalar",
+		in: `
+			x: #Def
+			x: 1
+			#Def: int
+			`,
+		sel:   cue.AnyDefinition,
+		allow: true,
 	}, {
 		desc: "allow field in any",
 		in: `
@@ -1622,8 +1748,6 @@ func TestAllows(t *testing.T) {
 		`,
 		sel:   cue.AnyString,
 		allow: true,
-
-		todo_nosharing: true,
 	}, {
 		desc: "disallow label in disjunction",
 		in: `
@@ -1666,9 +1790,6 @@ func TestAllows(t *testing.T) {
 
 	for _, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, tc.desc, func(t *testing.T, m *cuetdtest.M) {
-			if tc.todo_nosharing {
-				m.TODO_NoSharing(t)
-			}
 			ctx := m.CueContext()
 			v := mustCompile(t, ctx, tc.in)
 			v = v.LookupPath(path)
@@ -1747,10 +1868,11 @@ func TestValue_LookupDef(t *testing.T) {
 // TODO: trim down to individual defaults?
 func TestDefaults(t *testing.T) {
 	testCases := []struct {
-		value string
-		def   string
-		val   string
-		ok    bool
+		value      string
+		def        string
+		val        string
+		ok         bool
+		firstField bool
 	}{{
 		value: `number | *1`,
 		def:   "1",
@@ -1776,12 +1898,36 @@ func TestDefaults(t *testing.T) {
 		def:   `"x" | string`,
 		val:   `"x"|string`,
 		ok:    false,
+	}, {
+		value: `{#x: {y: #y}, #y: {a: int}}.#y`,
+		def:   `{a:int}`,
+		val:   `{a: int}`,
+		ok:    false,
+	}, {
+		value:      `{x?: []}`,
+		def:        `[]`,
+		val:        `[]`,
+		ok:         true,
+		firstField: true,
 	}}
 	for _, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, tc.value, func(t *testing.T, m *cuetdtest.M) {
 			v := getValue(m, "a: "+tc.value).Lookup("a")
 
 			v = v.Eval()
+
+			if tc.firstField {
+				iter, err := v.Fields(cue.All())
+				if err != nil {
+					t.Fatal(err)
+				}
+				if iter.Next() {
+					v = iter.Value()
+				} else {
+					t.Fatal("No fields found")
+				}
+			}
+
 			d, ok := v.Default()
 			if ok != tc.ok {
 				t.Errorf("hasDefault: got %v; want %v", ok, tc.ok)
@@ -1847,8 +1993,6 @@ func TestTemplate(t *testing.T) {
 		value string
 		path  []string
 		want  string
-
-		skip bool
 	}{{
 		value: `
 		a: [Name=string]: Name
@@ -1880,15 +2024,9 @@ func TestTemplate(t *testing.T) {
 		`,
 		path: []string{"a", "foo", "b", ""},
 		want: `{"c":"foolabel","d":"label"}`,
-
-		skip: true, // TODO: reordering
 	}}
 	for _, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, "", func(t *testing.T, m *cuetdtest.M) {
-			if tc.skip {
-				m.TODO_V3(t) // P2: reordering
-			}
-
 			v := getValue(m, tc.value)
 			for _, p := range tc.path {
 				if p == "" {
@@ -1913,7 +2051,6 @@ func TestElem(t *testing.T) {
 		value string
 		path  []string
 		want  string
-		skip  bool
 	}{{
 		value: `
 		a: [...int]
@@ -1945,14 +2082,9 @@ func TestElem(t *testing.T) {
 		`,
 		path: []string{"a", "foo", "b", ""},
 		want: "{\n\tc: \"foo\" + string\n\td: string\n}",
-		skip: true, // TODO(p3): Skip because this is just a reordering.
 	}}
 	for _, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, "", func(t *testing.T, m *cuetdtest.M) {
-			if tc.skip {
-				m.TODO_V3(t) // P2: reordering
-			}
-
 			v := getValue(m, tc.value)
 			cue.ValueVertex(v).Finalize(cue.ValueCtx(v))
 			for _, p := range tc.path {
@@ -2109,6 +2241,7 @@ func TestUnify(t *testing.T) {
 		value string
 		pathA string
 		pathB string
+		pathC string
 		want  string
 	}
 	testCases := []testCase{{
@@ -2159,14 +2292,85 @@ func TestUnify(t *testing.T) {
 		pathA: a,
 		pathB: "#B",
 		want:  `{}`,
+	}, {
+		value: `
+			a: obj: initialField: "foo"
+			a: #x
+			#x: obj?: _
+			b: obj: extraField: "bar"
+			`,
+		pathA: a,
+		pathB: b,
+		want:  `{"obj":{"initialField":"foo","extraField":"bar"}}`,
+	}, {
+		value: `
+			a: obj: initialField: "foo"
+			a: #x
+			#x: obj?: {...}
+			b: obj: extraField: "bar"
+				`,
+		pathA: a,
+		pathB: b,
+		want:  `{"obj":{"initialField":"foo","extraField":"bar"}}`,
+	}, {
+		value: `
+			a: obj: initialField: "foo"
+			#x: obj?: _
+			b: obj: extraField: "bar"
+			`,
+		pathA: a,
+		pathB: "#x",
+		pathC: b,
+		want:  `{"obj":{"initialField":"foo","extraField":"bar"}}`,
+	}, {
+		value: `
+			a: obj: initialField: "foo"
+			#x: obj?: {...}
+			a: #x
+			b: extraField: "bar"
+			`,
+		pathA: "a.obj",
+		pathB: b,
+		want:  `{"initialField":"foo","extraField":"bar"}`,
+	}, {
+		// Issue 3706
+		value: `
+            #site: {
+                bar?: _kind & {
+                    kind!:       "bar1"
+                    bar1field?: string
+                }
+                _kind: kind!: string
+            }
+            b: {
+                "bar": {
+                    "kind": "bar1",
+                    "bar1field": "expr"
+                }
+            }
+        `,
+		pathA: "#site",
+		pathB: b,
+		want:  `{"bar":{"kind":"bar1","bar1field":"expr"}}`,
 	}}
+
+	matrix := cuetdtest.FullMatrix
+	// adt.DebugDeps = true
+	// adt.OpenGraphs = true
+	// matrix[0].Flags.LogEval = 1
+
 	// TODO(tdtest): use cuetest.Run when supported.
-	cuetdtest.FullMatrix.Do(t, func(t *testing.T, m *cuetdtest.M) {
+	matrix.Do(t, func(t *testing.T, m *cuetdtest.M) {
 		tdtest.Run(t, testCases, func(t *cuetest.T, tc *testCase) {
 			v := getValue(m, tc.value)
 			x := v.LookupPath(cue.ParsePath(tc.pathA))
 			y := v.LookupPath(cue.ParsePath(tc.pathB))
-			b, err := x.Unify(y).MarshalJSON()
+			x = x.Unify(y)
+			if tc.pathC != "" {
+				z := v.LookupPath(cue.ParsePath(tc.pathC))
+				x = x.Unify(z)
+			}
+			b, err := x.MarshalJSON()
 			if err != nil {
 				t.Fatal(err)
 			}
@@ -2175,6 +2379,69 @@ func TestUnify(t *testing.T) {
 	})
 }
 
+// TestUnify2 is similar to TestUnify, but uses CompileString and Validate.
+func TestUnify2(t *testing.T) {
+	type testCase struct {
+		a   string
+		b   string
+		err bool
+	}
+	testCases := []testCase{{
+		a:   `null | close({})`,
+		b:   `quux: "boom"`,
+		err: true,
+	}, {
+		a:   `close({[=~"a"]: _})`,
+		b:   `a: 1`,
+		err: false,
+	}, {
+		a:   `close({{[=~"a"]: _}})`,
+		b:   `a: 1`,
+		err: false,
+	}, {
+		a:   `close({{["d"]: _}})`,
+		b:   `e: 1`,
+		err: true,
+	}, {
+		a:   `close({{["d"]: f: int}})`,
+		b:   `d: e: 1`,
+		err: false,
+	}, {
+		a:   `#D, #D: d: f: int`,
+		b:   `d: e: 1`,
+		err: true,
+	}, {
+		a:   `_x, _x: null | close({ foo?: bool })`,
+		b:   `bar: false`,
+		err: true,
+	}, {
+		a:   `_x, _x: null | close({ foo?: _x })`,
+		b:   `bar: false`,
+		err: true,
+	}, {
+		a:   `#d, #d: null | { foo?: _x }`,
+		b:   `bar: false`,
+		err: true,
+	}}
+
+	cuetdtest.Run(t, testCases, func(t *cuetdtest.T, tc *testCase) {
+		ctx := t.M.CueContext()
+
+		a := ctx.CompileString(tc.a)
+		b := ctx.CompileString(tc.b)
+
+		// r := (*runtime.Runtime)(ctx)
+		// r.SetDebugOptions(&cuedebug.Config{LogEval: 1})
+
+		c := a.Unify(b)
+
+		err := c.Validate()
+		t.Equal(err != nil, tc.err, "hasError")
+		t.Log("error:", err)
+	})
+
+}
+
 func TestUnifyAccept(t *testing.T) {
 	type testCase struct {
 		value string
@@ -2188,13 +2455,19 @@ func TestUnifyAccept(t *testing.T) {
 		want:  `"foo"`,
 	}, {
 		value: `#v: {a: "foo"}, #w: {b: 4}, #accept: {a: string, b: int}`,
-		want:  `{"a":"foo","b":4}`,
+		want: `{
+	a: "foo"
+	b: 4
+}`,
 	}, {
 		value: `#v: [string,  4], #w: ["foo", 4], #accept: [string, int, ...]`,
-		want:  `["foo",4]`,
+		want:  `["foo", 4]`,
 	}, {
 		value: `#v: {a: string, b: 1, _#hidden: int}, #w: {a: "foo"}, #accept: {...}`,
-		want:  `{"a":"foo","b":1}`,
+		want: `{
+	a: "foo"
+	b: 1
+}`,
 	}, {
 		// Issue #2325: let should not result in a closedness error.
 		value: `#accept: {
@@ -2212,24 +2485,147 @@ func TestUnifyAccept(t *testing.T) {
 		#w: {b:1}
 		#accept: {...}
 		`,
-		want: `{"b":1}`,
+		want: `{
+	b: 1
+}`,
 	}}
+
+	matrix := cuetdtest.FullMatrix
+	// adt.OpenGraphs = true
+	// matrix[0].Flags.LogEval = 1
 	// TODO(tdtest): use cuetest.Run when supported.
-	cuetdtest.FullMatrix.Do(t, func(t *testing.T, m *cuetdtest.M) {
+
+	matrix.Do(t, func(t *testing.T, m *cuetdtest.M) {
 		tdtest.Run(t, testCases, func(t *cuetest.T, tc *testCase) {
 			v := getValue(m, tc.value)
+			if err := v.Err(); err != nil {
+				t.Fatal(err)
+			}
 			x := v.LookupPath(cue.ParsePath("#v"))
 			y := v.LookupPath(cue.ParsePath("#w"))
 			a := v.LookupPath(cue.ParsePath("#accept"))
-			b, err := x.UnifyAccept(y, a).MarshalJSON()
-			if err != nil {
+			z := x.UnifyAccept(y, a)
+			if err := z.Err(); err != nil {
 				t.Fatal(err)
 			}
-			t.Equal(string(b), tc.want)
+			t.Equal(fmt.Sprint(z), tc.want)
 		})
 	})
 }
 
+// TestConjunctDedup tests a case where the same expression is added within
+// multiple Vertices. Normally, duplicate expressions happen when the same
+// Vertex is unified into the same node multiple times. But when CUE is
+// programmatically created, for instance through an API, the same expression
+// may end up in different vertices. Such duplicate expressions, when dedupped,
+// need to be handled similarly as when dedupping normal Vertices. This test
+// tests this case.
+//
+// Issue #3829
+func TestConjunctDedup(t *testing.T) {
+	type testCase struct {
+		value     string
+		want      string
+		conjuncts string // only tested if not ""
+	}
+	testCases := []testCase{{
+		value: `
+		#t: {} | {f1: int}
+		#t: {f: string}
+		`,
+		want: `{
+	f: string
+} | {
+	f1: int
+	f:  string
+}`,
+	}, {
+		// Issue #3847
+		value: `
+		import "struct"
+
+		#t: struct.MinFields(4)
+		#t: struct.MaxFields(9)
+		`,
+		want:      `struct.MinFields(4) & struct.MaxFields(9)`,
+		conjuncts: `[struct.MinFields(4) struct.MaxFields(9)]`,
+	}}
+
+	matrix := cuetdtest.FullMatrix
+	matrix.Do(t, func(t *testing.T, m *cuetdtest.M) {
+		tdtest.Run(t, testCases, func(t *cuetest.T, tc *testCase) {
+			ctx := m.CueContext()
+			tPath := cue.MakePath(cue.Def("#t"))
+			vOrig := ctx.CompileString(tc.value).LookupPath(tPath)
+
+			if err := vOrig.Err(); err != nil {
+				t.Fatal(err)
+			}
+
+			_, args := vOrig.Expr()
+
+			var v cue.Value
+			for _, a := range args {
+				v = v.UnifyAccept(a, vOrig)
+				if err := v.Err(); err != nil {
+					t.Fatal(errors.Details(v.Err(), nil))
+				}
+			}
+			t.Equal(fmt.Sprint(v), tc.want)
+
+			_, args = v.Expr()
+			if tc.conjuncts == "" {
+				return
+			}
+			t.Equal(fmt.Sprint(args), tc.conjuncts)
+		})
+	})
+}
+
+// TestIssue3826 tests that if the same value gets added through multiple paths,
+// as can happen more readily through the API, that the deduplication of the
+// conjuncts works as expected.
+func TestIssue3826(t *testing.T) {
+	cuetdtest.FullMatrix.Run(t, "test", func(t *testing.T, m *cuetdtest.M) {
+		ctx := m.CueContext()
+
+		v := ctx.CompileString(`
+			_schema: _
+			_schema
+			apiVersion: "foo"
+			spec: group: "foo"
+		`)
+		schema := ctx.CompileString(`
+			#Schema: spec!: group!: string
+		`).LookupPath(cue.MakePath(cue.Def("#Schema")))
+
+		v = v.FillPath(cue.MakePath(cue.Hid("_schema", "_")), schema)
+
+		if err := v.Validate(); err != nil {
+			t.Error(err)
+		}
+	})
+}
+
+// TestIssue4093 tests that a value that's a *adt.NodeLink
+// under the hood can be formatted OK.
+func TestIssue4093(t *testing.T) {
+	cuetdtest.FullMatrix.Run(t, "test", func(t *testing.T, m *cuetdtest.M) {
+		v := m.CueContext().CompileString(`x: #X, #X: int`)
+		qt.Assert(t, qt.IsNil(v.Err()))
+		v = v.LookupPath(cue.MakePath(cue.Str("x")))
+		op, args := v.Expr()
+		qt.Assert(t, qt.Equals(op, cue.SelectorOp))
+		qt.Assert(t, qt.HasLen(args, 2))
+		// Note: we see the entire value because the #X is represented
+		// internally as a selector operation topLevelValue.#X
+		// and we're looking at the first argument to that operation.
+		// It's not clear why we see `x: int` not `x: #X`.
+		s := fmt.Sprintf("%#v", args[0])
+		qt.Assert(t, qt.Equals(s, "x:  int\n#X: int"))
+	})
+}
+
 func TestEquals(t *testing.T) {
 	testCases := []struct {
 		a, b string
@@ -2280,6 +2676,10 @@ func TestEquals(t *testing.T) {
 		`{ a: 2, b: { 3 } }`,
 		`{ a: { 2 }, b: 3 }`,
 		true,
+	}, {
+		`{a: *1 | 2 | 3}`,
+		`{a: 3 | *1}`,
+		true,
 	}}
 	for _, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, "", func(t *testing.T, m *cuetdtest.M) {
@@ -2302,8 +2702,6 @@ func TestValidate(t *testing.T) {
 		in   string
 		err  bool
 		opts []cue.Option
-
-		skip bool
 	}{{
 		desc: "issue #51",
 		in: `
@@ -2382,15 +2780,6 @@ func TestValidate(t *testing.T) {
 			`,
 		opts: []cue.Option{cue.DisallowCycles(true)},
 		err:  true,
-
-		// TODO: in the new evaluator these are not considered to be cycles
-		// but rather incomplete errors. This is actually the correct behavior
-		// and the old evaluator treats errors like this by default.
-		// The option tested in this test was added for backwards compatibility
-		// when the old evaluator was made to treat these kinds of errors
-		// equally. With the new evaluator we can no longer distinguish these
-		// errors. For now, at least. Consider what to do with this option.
-		skip: true,
 	}, {
 		desc: "builtins are okay",
 		in: `
@@ -2438,10 +2827,6 @@ func TestValidate(t *testing.T) {
 	}}
 	for _, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, tc.desc, func(t *testing.T, m *cuetdtest.M) {
-			if tc.skip {
-				m.TODO_V3(t) // P1: wrong results
-			}
-
 			ctx := m.CueContext()
 			val := ctx.CompileString(tc.in, cue.Filename("validate"))
 			err := val.Validate(tc.opts...)
@@ -2570,7 +2955,7 @@ func TestValueLookup(t *testing.T) {
 				t.Errorf("exists: got %v; want %v", got, tc.notExists)
 			}
 
-			got := cue.ValueCtx(v).Str(cue.ValueVertex(v))
+			got := cue.ValueCtx(v).String(cue.ValueVertex(v))
 			if tc.str == "" {
 				t.Fatalf("str empty, got %q", got)
 			}
@@ -2653,7 +3038,6 @@ func TestValueDoc(t *testing.T) {
 			val  cue.Value
 			path string
 			doc  string
-			skip bool
 		}{{
 			val:  v1,
 			path: "foos",
@@ -2685,12 +3069,10 @@ field1 is an int.
 		}, {
 			val:  v1,
 			path: "baz field1",
-			doc: `comment from bar on field 1
+			doc: `comment from baz on field 1
 
-comment from baz on field 1
+comment from bar on field 1
 `,
-			// New evaluaotor orders the comments differently (arguably better).
-			skip: true,
 		}, {
 			val:  v1,
 			path: "baz field2",
@@ -2709,9 +3091,6 @@ Another Foo.
 `,
 		}}
 		for _, tc := range testCases {
-			if tc.skip {
-				m.TODO_V3(t) // P2: reordering
-			}
 			t.Run("field:"+tc.path, func(t *testing.T) {
 				v := tc.val.Lookup(strings.Split(tc.path, " ")...)
 				doc := docStr(v.Doc())
@@ -2823,29 +3202,32 @@ func TestMarshalJSON(t *testing.T) {
 		value: `a: 1.0/1`,
 		json:  `{"a":1.0}`,
 	}, {
-		// Issue #108
+		// Issue #108: no longer supported.
 		value: `
 		a: int
-		a: >0
+		a: 1
 		a: <2
 
 		b: int
-		b: >=0.9
+		b: 1
 		b: <1.1
 
 		c: int
 		c: >1
-		c: <=2
+		c: 2
 
 		d: int
 		d: >=1
 		d: <=1.5
+		d: 1
 
 		e: int
 		e: >=1
 		e: <=1.32
+		e: 1
 
 		f: >=1.1 & <=1.1
+		f: 1.1
 		`,
 		json: `{"a":1,"b":1,"c":2,"d":1,"e":1,"f":1.1}`,
 	}, {
@@ -2866,7 +3248,7 @@ func TestMarshalJSON(t *testing.T) {
 	}, {
 		// Issue #326
 		value: `x: "\(string)": "v"`,
-		err:   `x: invalid interpolation`,
+		err:   `x: key value of dynamic field must be concrete, found _|_(x: invalid interpolation: non-concrete value string (type string)) (and 1 more errors)`,
 	}, {
 		// Issue #326
 		value: `x: "\(bool)": "v"`,
@@ -2909,8 +3291,6 @@ func TestMarshalJSON(t *testing.T) {
 	}}
 	for i, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, fmt.Sprintf("%d/%v", i, tc.value), func(t *testing.T, m *cuetdtest.M) {
-			m.TODO_V3(t) // P1: wrong output
-
 			val := getValue(m, tc.value)
 			b, err := val.MarshalJSON()
 			checkFatal(t, err, tc.err, "init")
@@ -2990,8 +3370,6 @@ func TestWalk(t *testing.T) {
 	}}
 	for i, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, fmt.Sprintf("%d/%v", i, tc.value), func(t *testing.T, m *cuetdtest.M) {
-			m.TODO_V3(t) // P1: missing undefined field
-
 			val := getValue(m, tc.value)
 			buf := []byte{}
 			stripComma := func() {
@@ -3175,7 +3553,6 @@ func TestPos(t *testing.T) {
 	testCases := []struct {
 		value string
 		pos   string
-		skip  bool
 	}{{
 		value: `
 a: string
@@ -3185,11 +3562,7 @@ a: "foo"`,
 		value: `
 a: x: string
 a: x: "x"`,
-		pos: "2:4",
-
-		// the position of the new evaluator is also correct, and actually
-		// better.
-		skip: true,
+		pos: "3:4",
 	}, {
 		// Prefer struct conjuncts with actual fields.
 		value: `
@@ -3209,10 +3582,6 @@ a: x: y: z: "x"`,
 	}}
 	for _, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, "", func(t *testing.T, m *cuetdtest.M) {
-			if tc.skip {
-				m.TODO_V3(t) // P3: different position
-			}
-
 			c := m.CueContext()
 			v := c.CompileString(tc.value)
 			v = v.LookupPath(cue.ParsePath("a"))
@@ -3229,7 +3598,6 @@ func TestPathCorrection(t *testing.T) {
 		input  string
 		lookup func(i cue.Value) cue.Value
 		want   string
-		skip   bool
 	}{{
 		input: `
 			a: b: {
@@ -3467,9 +3835,6 @@ func TestPathCorrection(t *testing.T) {
 		},
 	}}
 	for _, tc := range testCases {
-		if tc.skip {
-			continue
-		}
 		cuetdtest.FullMatrix.Run(t, "", func(t *testing.T, m *cuetdtest.M) {
 			ctx := m.CueContext()
 
@@ -3592,6 +3957,7 @@ func checkFailed(t *testing.T, err error, str, name string) bool {
 func TestExpr(t *testing.T) {
 	testCases := []struct {
 		input string
+		path  string // path to lookup, defaults to "v"
 		want  string
 	}{{
 		input: "v: 3",
@@ -3671,6 +4037,19 @@ func TestExpr(t *testing.T) {
 	}, {
 		input: "v: 2 mod 5",
 		want:  "mod(2 5)",
+	}, {
+		input: "@experiment(explicitopen)\n v: #Y..., #Y: {b: 2}",
+		want:  `...(.(〈〉 "#Y"))`,
+	}, {
+		input: "@experiment(explicitopen)\n v: {#Y..., a: 1}, #Y: {b: 2}",
+		want:  `&(...(.(〈〉 "#Y")) {a:1})`,
+	}, {
+		input: "@experiment(explicitopen)\n v: {#Y, a: 1}, #Y: {b: 2}",
+		want:  `&(.(〈〉 "#Y") {a:1})`,
+	}, {
+		// Note: before explicitopen there is no way to detect embeddings.
+		input: "v: {#Y, a: 1}, #Y: {b: 2}",
+		want:  `&(.(〈〉 "#Y") {a:1})`,
 	}, {
 		input: "v: a.b, a: b: 4",
 		want:  `.(.(〈〉 "a") "b")`,
@@ -3742,10 +4121,30 @@ func TestExpr(t *testing.T) {
 	}, {
 		input: `a: string, if true { v: a }`,
 		want:  `.(〈〉 "a")`,
+	}, {
+		// Issue #4198: panic with opID mismatch when calling Expr() on LookupPath result.
+		input: `
+			x: #B & {
+				y: [ 1 ]
+			}
+
+			#B: #A | []
+
+			#A: {
+				y?: [...int]
+				...
+			}
+		`,
+		path: "x.y",
+		want: `&([1] [...int])`,
 	}}
 	for _, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, tc.input, func(t *testing.T, m *cuetdtest.M) {
-			v := getValue(m, tc.input).Lookup("v")
+			path := tc.path
+			if path == "" {
+				path = "v"
+			}
+			v := getValue(m, tc.input).LookupPath(cue.ParsePath(path))
 			got := exprStr(v)
 			if got != tc.want {
 				t.Errorf("\n got %v;\nwant %v", got, tc.want)
@@ -3776,3 +4175,300 @@ func compactRawStr(v cue.Value) string {
 	cfg := &debug.Config{Compact: true, Raw: true}
 	return debug.NodeString(ctx, cue.ValueVertex(v), cfg)
 }
+
+func TestIsClosed(t *testing.T) {
+	testCases := []struct {
+		desc                string
+		input               string
+		path                string
+		wantClosed          bool
+		wantClosedRecursive bool
+	}{
+		{
+			desc:                "open struct",
+			input:               `x: {a: 1}`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "struct with close builtin",
+			input:               `x: close({a: 1})`,
+			path:                "x",
+			wantClosed:          true,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "definition",
+			input:               `#X: {a: 1}`,
+			path:                "#X",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "field in definition",
+			input:               `#X: {a: {b: 1}}`,
+			path:                "#X.a",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "structure sharing: definition reference",
+			input:               `#X: {a: 1}, y: #X`,
+			path:                "y",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "structure sharing: definition nested field",
+			input:               `#X: {a: {b: 1}}, y: #X`,
+			path:                "y.a",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "structure sharing: mixed open and closed",
+			input:               `#X: {a: 1}, y: {b: 2, c: #X}`,
+			path:                "y.c",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "structure sharing: open struct in closed",
+			input:               `#X: {a: open}, open: {b: 1}, y: #X.a`,
+			path:                "y",
+			wantClosed:          true, // Inherits closed status from definition
+			wantClosedRecursive: true, // Inherits closed status from definition
+		},
+		{
+			desc:                "disjunction with close",
+			input:               `x: close({a: 1}) | close({b: 2})`,
+			path:                "x",
+			wantClosed:          false, // Disjunction itself is not closed
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "disjunction with close and elimination",
+			input:               `x: close({a: 1}) | close({b: 2}), x: a: 1`,
+			path:                "x",
+			wantClosed:          true,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "disjunction with definition",
+			input:               `#A: {a: 1}, #B: {b: 2}, x: #A | #B`,
+			path:                "x",
+			wantClosed:          false, // Disjunction itself is not closed
+			wantClosedRecursive: false, // Disjunction itself is not closed
+		},
+		{
+			desc:                "disjunction with mixed closedness",
+			input:               `#A: {a: 1}, x: #A | {b: 2}`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "disjunction resolved to closed",
+			input:               `#A: {a: 1}, #B: {b: 2}, x: (#A | #B) & {a: 1}`,
+			path:                "x",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "nested definition in open struct",
+			input:               `x: {#D: {a: 1}}`,
+			path:                "x.#D",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "empty struct with close",
+			input:               `x: close({})`,
+			path:                "x",
+			wantClosed:          true,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "empty definition",
+			input:               `#X: {}`,
+			path:                "#X",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "close with pattern constraint",
+			input:               `x: close({[string]: int})`,
+			path:                "x",
+			wantClosed:          true,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "definition with pattern constraint",
+			input:               `#X: {[string]: int}`,
+			path:                "#X",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "multiple levels of structure sharing 1",
+			input:               `#A: {a: 1}, #B: {b: #A}, y: #B.b`,
+			path:                "y",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "multiple levels of structure sharing 2",
+			input:               `#A: {a: 1}, x: {b: #A}, y: x.b`,
+			path:                "y",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "multiple levels of structure sharing 3",
+			input:               `#A: {a: {c: 1}}, x: {b: #A}, y: x.b.a`,
+			path:                "y",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "multiple levels of structure sharing 4",
+			input:               `a: {a: c: 1}, x: {#b: a}, y: x.#b.a`,
+			path:                "y",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "scalar value (non-struct)",
+			input:               `x: 42`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "list value",
+			input:               `x: [1, 2, 3]`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc: "unresolved disjunction with all closed options",
+			input: `
+				#A: {a: 1}
+				#B: {b: 2}
+				x: #A | #B
+			`,
+			path:                "x",
+			wantClosed:          false, // Disjunction itself is not closed
+			wantClosedRecursive: false,
+		},
+		{
+			desc: "unresolved disjunction with mixed closedness",
+			input: `
+				#A: {a: 1}
+				x: #A | {b: 2}
+			`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc: "disjunction with close() builtin",
+			input: `
+				x: close({a: 1}) | close({b: 2})
+			`,
+			path:                "x",
+			wantClosed:          false, // Disjunction itself is not closed
+			wantClosedRecursive: false,
+		},
+		{
+			desc: "default disjunction with closed values",
+			input: `
+				#A: {a: 1}
+				#B: {b: 2}
+				x: *#A | #B
+			`,
+			path:                "x",
+			wantClosed:          false, // Disjunction itself is not closed
+			wantClosedRecursive: false,
+		},
+		{
+			desc: "nested disjunction in definition",
+			input: `
+				x: #X
+				#X: {
+					field: close({a: 1}) | close({b: 2})
+				}
+			`,
+			path:                "x",
+			wantClosed:          true, // Definition is closed
+			wantClosedRecursive: true, // Definition is recursively closed
+		},
+		{
+			desc: "triple disjunction all closed",
+			input: `
+				#A: {a: 1}
+				#B: {b: 2}
+				#C: {c: 3}
+				x: #A | #B | #C
+			`,
+			path:                "x",
+			wantClosed:          false, // Disjunction itself is not closed
+			wantClosedRecursive: false,
+		},
+		{
+			desc: "triple disjunction with one open",
+			input: `
+				#A: {a: 1}
+				#B: {b: 2}
+				x: #A | #B | {c: 3}
+			`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "definition with disjunction of scalar and struct",
+			input:               `#A: 1 | {a!: int}`,
+			path:                "#A",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "closed list with definition elements",
+			input:               `#A: {a: 1}, x: [#A, #A]`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "list element from definition",
+			input:               `#A: {a: 1}, x: [#A, #A]`,
+			path:                "x[0]",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+	}
+
+	for _, tc := range testCases {
+		t.Run(tc.desc, func(t *testing.T) {
+			cuetdtest.FullMatrix.Run(t, tc.input, func(t *testing.T, m *cuetdtest.M) {
+				v := getValue(m, tc.input)
+
+				v = v.LookupPath(cue.ParsePath(tc.path))
+
+				if err := v.Err(); err != nil {
+					t.Fatalf("unexpected error: %v", err)
+				}
+
+				qt.Check(t, qt.Equals(v.IsClosed(), tc.wantClosed))
+				qt.Check(t, qt.Equals(v.IsClosedRecursively(), tc.wantClosedRecursive))
+
+				if v.IsClosedRecursively() {
+					qt.Check(t, qt.IsTrue(v.IsClosed()))
+				}
+			})
+		})
+	}
+}
