@@ -21,6 +21,7 @@ package debug
 
 import (
 	"fmt"
+	"slices"
 	"strconv"
 	"strings"
 
@@ -38,6 +39,11 @@ type Config struct {
 	Cwd     string
 	Compact bool
 	Raw     bool
+
+	// ExpandLetExpr causes the expression of let reference to be printed.
+	// Note that this may result in large outputs. Use with care.
+	// Only applies if Compact is false.
+	ExpandLetExpr bool
 }
 
 // AppendNode writes a string representation of the node to w.
@@ -45,12 +51,7 @@ func AppendNode(dst []byte, i adt.StringIndexer, n adt.Node, config *Config) []b
 	if config == nil {
 		config = &Config{}
 	}
-	p := printer{dst: dst, index: i, cfg: config}
-	if config.Compact {
-		p := compactPrinter{p}
-		p.node(n)
-		return p.dst
-	}
+	p := printer{dst: dst, index: i, cfg: config, compact: config.Compact}
 	p.node(n)
 	return p.dst
 }
@@ -65,10 +66,11 @@ func NodeString(i adt.StringIndexer, n adt.Node, config *Config) string {
 }
 
 type printer struct {
-	dst    []byte
-	index  adt.StringIndexer
-	indent string
-	cfg    *Config
+	dst     []byte
+	index   adt.StringIndexer
+	indent  string
+	cfg     *Config
+	compact bool // copied from config.Compact
 
 	// keep track of vertices to avoid cycles.
 	stack []*adt.Vertex
@@ -80,8 +82,51 @@ type printer struct {
 	// - auto
 }
 
+// ReplaceArg implements the format.Printer interface. It wraps Vertex arguments
+// with a formatter value, that holds a pointer to w. This allows the stack
+// of processed vertices to be passed down, which in turn is used for cycle
+// detection.
+func (w *printer) ReplaceArg(arg any) (replacement any, replaced bool) {
+	var x adt.Node
+	var r adt.Runtime
+	switch v := arg.(type) {
+	case adt.Node:
+		x = v
+	case adt.Formatter:
+		x = v.X
+		r = v.R
+	}
+
+	switch x := x.(type) {
+	default:
+		return arg, false
+	case *adt.Vertex:
+		// We replace the formatter (or node) with our own formatter that is
+		// capable of detecting cycles.
+		return formatter{p: w, x: x, r: r}, true
+	}
+}
+
+type formatter struct {
+	p *printer
+	x adt.Node
+	r adt.Runtime ``
+}
+
+func (f formatter) String() string {
+	p := printer{
+		dst:     make([]byte, 0, 128),
+		index:   f.r,
+		cfg:     f.p.cfg,
+		compact: true, // Always compact for error arguments.
+		stack:   f.p.stack,
+	}
+	p.node(f.x)
+	return string(p.dst)
+}
+
 func (w *printer) string(s string) {
-	if len(w.indent) > 0 {
+	if !w.compact && len(w.indent) > 0 {
 		s = strings.Replace(s, "\n", "\n"+w.indent, -1)
 	}
 	w.dst = append(w.dst, s...)
@@ -132,19 +177,15 @@ func (w *printer) shared(v *adt.Vertex) {
 // of v, if it is a shared node. It reports the dereferenced node and whether
 // the node was printed.
 func (w *printer) printShared(v0 *adt.Vertex) (x *adt.Vertex, ok bool) {
-
 	// Handle cyclic shared nodes differently.  If a shared node was part of
 	// a disjunction, it will still be wrapped in a disjunct Vertex.
 	// Similarly, a shared node should never point to a disjunct directly,
 	// but rather to the original arc that subsequently points to a
 	// disjunct.
 	v0 = v0.DerefDisjunct()
-	isCyclic := v0.IsCyclic
 	s, ok := v0.BaseValue.(*adt.Vertex)
 	v1 := v0.DerefValue()
-	useReference := v0.IsShared && !v1.IsDetached() && v1.MayAttach()
-	isCyclic = isCyclic || v1.IsCyclic
-	_ = isCyclic
+	useReference := v0.IsShared && !v1.Internal()
 	// NOTE(debug): use this line instead of the following to expand shared
 	// cases where it is safe to do so.
 	// if useReference && isCyclic && ok && len(v.Arcs) > 0 {
@@ -164,11 +205,11 @@ func (w *printer) printShared(v0 *adt.Vertex) (x *adt.Vertex, ok bool) {
 }
 
 func (w *printer) pushVertex(v *adt.Vertex) bool {
-	for _, x := range w.stack {
-		if x == v {
-			w.string("<TODO: unmarked structural cycle>")
-			return false
-		}
+	if slices.Contains(w.stack, v) {
+		w.string("value at path '")
+		w.path(v)
+		w.string("'")
+		return false
 	}
 	w.stack = append(w.stack, v)
 	return true
@@ -178,21 +219,15 @@ func (w *printer) popVertex() {
 	w.stack = w.stack[:len(w.stack)-1]
 }
 
-func (w *printer) shortError(errs errors.Error) {
-	for {
-		msg, args := errs.Msg()
-		w.dst = fmt.Appendf(w.dst, msg, args...)
-
-		err := errors.Unwrap(errs)
-		if err == nil {
-			break
-		}
-
-		if errs, _ = err.(errors.Error); errs != nil {
-			w.string(err.Error())
-			break
-		}
-	}
+// TODO: always print path? We allow a choice for keeping the error diff at a
+// minimum.
+func (w *printer) shortError(errs errors.Error, omitPath bool) {
+	w.string(errors.StringWithConfig(errs, &errors.Config{
+		Cwd:      w.cfg.Cwd,
+		ToSlash:  true,
+		OmitPath: omitPath,
+		Printer:  w,
+	}))
 }
 
 func (w *printer) interpolation(x *adt.Interpolation) {
@@ -225,7 +260,21 @@ func (w *printer) interpolation(x *adt.Interpolation) {
 	w.string(quote)
 }
 
+func (w *printer) arg(n adt.Node) {
+	if x, ok := n.(*adt.Vertex); ok {
+		if x.Label != adt.InvalidLabel {
+			w.path(x)
+			return
+		}
+	}
+	w.node(n)
+}
+
 func (w *printer) node(n adt.Node) {
+	if w.compact {
+		w.compactNode(n)
+		return
+	}
 	switch x := n.(type) {
 	case *adt.Vertex:
 		x, ok := w.printShared(x)
@@ -262,10 +311,11 @@ func (w *printer) node(n adt.Node) {
 			w.indent += "// "
 			w.string("\n")
 			w.dst = fmt.Appendf(w.dst, "[%v]", v.Code)
-			if !v.ChildError {
+			if !v.ChildError || len(x.Arcs) == 0 {
 				msg := errors.Details(v.Err, &errors.Config{
 					Cwd:     w.cfg.Cwd,
 					ToSlash: true,
+					Printer: w,
 				})
 				msg = strings.TrimSpace(msg)
 				if msg != "" {
@@ -427,7 +477,7 @@ func (w *printer) node(n adt.Node) {
 		w.string(`_|_`)
 		if x.Err != nil {
 			w.string("(")
-			w.shortError(x.Err)
+			w.shortError(x.Err, true)
 			w.string(")")
 		}
 
@@ -507,6 +557,10 @@ func (w *printer) node(n adt.Node) {
 		w.string(";let ")
 		w.label(x.Label)
 		w.string(closeTuple)
+		if w.cfg.ExpandLetExpr {
+			w.string("=>")
+			w.node(x.X)
+		}
 
 	case *adt.SelectorExpr:
 		w.node(x.X)
@@ -551,6 +605,10 @@ func (w *printer) node(n adt.Node) {
 		w.node(x.Y)
 		w.string(")")
 
+	case *adt.OpenExpr:
+		w.node(x.X)
+		w.string("...")
+
 	case *adt.CallExpr:
 		w.node(x.Fun)
 		w.string("(")
@@ -558,7 +616,7 @@ func (w *printer) node(n adt.Node) {
 			if i > 0 {
 				w.string(", ")
 			}
-			w.node(a)
+			w.arg(a)
 		}
 		w.string(")")
 
@@ -576,7 +634,7 @@ func (w *printer) node(n adt.Node) {
 			if i > 0 {
 				w.string(", ")
 			}
-			w.node(a)
+			w.arg(a)
 		}
 		w.string(")")
 
